# uncompyle6 version 3.8.1.dev0
# Python bytecode 3.8.0 (3413)
# Decompiled from: Python 3.8.11 (default, Aug  3 2021, 15:09:35)
# [GCC 7.5.0]
# Embedded file name: /data/pinello/PROJECTS/2021_08_ANBE/software/bean/bean/framework/ReporterScreen.py
# Compiled at: 2022-01-06 10:00:30
pass
# okay decompiling ReporterScreen.cpython-38.pyc
